"""
Variable definitions for Hypernets land a water network data products
"""
from copy import deepcopy
from typing import Any

import numpy as np

# Sets variables are defined in the below dictionaries. Each entry in a dictionary defines a product variable, the value
# of each entry defines the variable attributes with a dictionary that must contain the following entries:
#
# * "dim" - list of variable dimensions, selected from defined dimension names
# * "dtype" - dtype of data as numpy type
# * "attributes" - dictionary of variable attributes
# * "encoding" - dictionary of variable encoding parameters
#
# Following the variable set definitions a dictionary of is built with an entry for each Hypernets product type, with
# entry of its variables built by combining the variables from the variable sets. This is called VARIABLES_DICT_DEFS
#
# Product variables are defined following the specification defined in:
# Hypernets Team, Product Data Format Specification, v0.5 (2020)


# Dimensions
WL_DIM = "wavelength"
CD_DIM = "calibrationdates"
NL_DIM = "nonlinearcoef"
ND_DIM = "nonlineardates"
WC_DIM = "wavcoef"
WD_DIM = "wavdates"

SERIES_DIM = "series"
SCAN_DIM = "scan"
Lu_SCAN_DIM = "scan"
SEQ_DIM = "sequence"

# Contributing variable sets
# --------------------------

# > COMMON_VARIABLES - Variables are required for a data products
COMMON_VARIABLES_SERIES = {
    "wavelength": {
        "dim": [WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "wavelength",
            "long_name": "Wavelength",
            "units": "nm",
            "preferred_symbol": "wv",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.025, "offset": 320},
    },
    "bandwidth": {
        "dim": [WL_DIM],
        "dtype": np.float32,
        "attributes": {},
        "encoding": {"dtype": np.uint16, "scale_factor": 0.1, "offset": 0.0},
    },
    "viewing_azimuth_angle": {
        "dim": [SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "sensor_azimuth_angle",
            "long_name": "sensor_azimuth_angle is the "
            "horizontal angle between the line of "
            "sight from the observation point to "
            "the sensor and a reference direction "
            "at the observation point, which is "
            "often due north. The angle is "
            "measured clockwise positive, starting"
            " from the reference direction. A "
            "comment attribute should be added to "
            "a data variable with this standard "
            "name to specify the reference "
            "direction.",
            "units": "degrees",
            "reference": "True North",
            "preferred_symbol": "vaa",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0055, "offset": 0.0},
    },
    "viewing_zenith_angle": {
        "dim": [SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "sensor_zenith_angle",
            "long_name": "sensor_zenith_angle is the angle "
            "between the line of sight to the "
            "sensor and the local zenith at the "
            "observation target. This angle is "
            "measured starting from directly "
            "overhead and its range is from zero "
            "(directly overhead the observation "
            "target) to 180 degrees (directly below"
            " the observation target). Local zenith"
            " is a line perpendicular to the "
            "Earth's surface at a given location. "
            "'Observation target' means a location"
            " on the Earth defined by the sensor "
            "performing the observations.",
            "units": "degrees",
            "preferred_symbol": "vza",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0028, "offset": 0.0},
    },
    "solar_azimuth_angle": {
        "dim": [SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "solar_azimuth_angle",
            "long_name": "Solar azimuth angle is the horizontal "
            "angle between the line of sight to the "
            "sun and a reference direction which "
            "is often due north. The angle is "
            "measured clockwise.",
            "units": "degrees",
            "reference": "True North",
            "preferred_symbol": "saa",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0055, "offset": 0.0},
    },
    "solar_zenith_angle": {
        "dim": [SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "solar_zenith_angle",
            "long_name": "Solar zenith angle is the the angle "
            "between the line of sight to the sun and"
            " the local vertical.",
            "units": "degrees",
            "preferred_symbol": "sza",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0028, "offset": 0.0},
    },
    "acquisition_time": {
        "dim": [SERIES_DIM],
        "dtype": np.uint32,
        "attributes": {
            "standard_name": "time",
            "long_name": "Acquisition time in seconds since " "1970-01-01 " "00:00:00",
            "units": "s",
        },
    },
    "series_id": {
        "dim": [SERIES_DIM],
        "dtype": np.uint8,
        "attributes": {
            "standard_name": "series_id",
            "long_name": "Series id number",
            "units": "-",
        },
    },
}

COMMON_VARIABLES_SEQ = deepcopy(COMMON_VARIABLES_SERIES)
COMMON_VARIABLES_SCAN = deepcopy(COMMON_VARIABLES_SERIES)
COMMON_VARIABLES_Lu_SCAN = deepcopy(COMMON_VARIABLES_SERIES)

for variable in COMMON_VARIABLES_SEQ.keys():
    COMMON_VARIABLES_SEQ[variable]["dim"] = [
        d if d != SERIES_DIM else SEQ_DIM
        for d in COMMON_VARIABLES_SERIES[variable]["dim"]
    ]
for variable in COMMON_VARIABLES_SCAN.keys():
    COMMON_VARIABLES_SCAN[variable]["dim"] = [
        d if d != SERIES_DIM else SCAN_DIM
        for d in COMMON_VARIABLES_SERIES[variable]["dim"]
    ]
for variable in COMMON_VARIABLES_Lu_SCAN.keys():
    COMMON_VARIABLES_Lu_SCAN[variable]["dim"] = [
        d if d != SERIES_DIM else Lu_SCAN_DIM
        for d in COMMON_VARIABLES_SERIES[variable]["dim"]
    ]


L0_RAD_VARIABLES = {
    "integration_time": {
        "dim": [SCAN_DIM],
        "dtype": np.uint32,
        "attributes": {
            "standard_name": "integration_time",
            "long_name": "Integration time during measurement",
            "units": "s",
        },
        "encoding": {"dtype": np.uint32},
    },
    "temperature": {
        "dim": [SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "temperature",
            "long_name": "temperature of instrument",
            "units": "degrees",
        },
        "encoding": {"dtype": np.float32},
    },
    "acceleration_x_mean": {
        "dim": [SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "acceleration_x_mean",
            "long_name": "Time during measurement",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "acceleration_x_std": {
        "dim": [SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "acceleration_x_std",
            "long_name": "",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "acceleration_y_mean": {
        "dim": [SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "acceleration_y_mean",
            "long_name": "",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "acceleration_y_std": {
        "dim": [SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "acceleration_y_std",
            "long_name": "",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "acceleration_z_mean": {
        "dim": [SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "acceleration_z_mean",
            "long_name": "",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "acceleration_z_std": {
        "dim": [SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "acceleration_z_std",
            "long_name": "",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "digital_number": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint32,
        "attributes": {
            "standard_name": "digital_number",
            "long_name": "Digital number, raw data",
            "units": "-",
            "unc_comps": ["u_rel_random_digital_number"],
        },
    },
    "u_rel_random_digital_number": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "random relative uncertainty on digital number",
            "long_name": "random relative uncertainty on digital number",
            "units": "-",
            "err_corr": [
                {"dim": SCAN_DIM, "form": "random", "params": [], "units": []},
                {"dim": WL_DIM, "form": "random", "params": [], "units": []},
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    # "series_id": {"dim": [SCAN_DIM],
    #                "dtype": np.chararray,
    #                "attributes": {"standard_name": "series_id",
    #                               "long_name": "series_id",
    #                               "units": "-"}}
    "dark_signal": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint32,
        "attributes": {
            "standard_name": "dark_digital_number",
            "long_name": "Digital number, raw data for dark signal",
            "units": "-",
        },
    },
}

L0_IRR_VARIABLES = L0_RAD_VARIABLES
L0_BLA_VARIABLES = L0_RAD_VARIABLES

CAL_VARIABLES = {
    "wavelengths": {
        "dim": [WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "wavelengths",
            "long_name": "Wavelengths",
            "units": "nm",
            "preferred_symbol": "wv",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.025, "offset": 320},
    },
    "wavpix": {
        "dim": [WL_DIM],
        "dtype": np.uint16,
        "attributes": {
            "standard_name": "wavelength pixel",
            "long_name": "Wavelength pixel in L0 data that corresponds to the given wavelength",
        },
    },
    "gains": {
        "dim": [WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "gains",
            "long_name": "gains (calibration coefficients)",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_systematic_indep_gains",
                "u_rel_systematic_corr_rad_irr_gains",
            ],
        },
    },
    "u_rel_systematic_indep_gains": {
        "dim": [WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on gains",
            "long_name": "the systematic relative uncertainty component on gains (calibration coefficients) that is not correlated between radiance and irradiance, divided by the gains",
            "units": "-",
            "err_corr": [
                {
                    "dim": WL_DIM,
                    "form": "err_corr_matrix",
                    "params": ["corr_systematic_indep_gains"],
                    "units": [],
                },
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_gains": {
        "dim": [WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "systematic relative uncertainty on gains (correlated radiance and irradiance)",
            "long_name": "the systematic relative uncertainty component on gains (calibration coefficients) that is correlated between radiance and irradiance",
            "units": "-",
            "err_corr": [
                {
                    "dim": WL_DIM,
                    "form": "err_corr_matrix",
                    "params": ["corr_systematic_corr_rad_irr_gains"],
                    "units": [],
                },
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_indep_gains": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on gains",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on gains (calibration coefficients) that is not correlated between radiance and irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_gains": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on gains (correlated radiance and irradiance)",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on gains (calibration coefficients) that is correlated between radiance and irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "wavelength_coefficients": {
        "dim": [WC_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "wavelength coefficients",
            "long_name": "Polynomial coefficients for wavelength scale",
            "units": "-",
        },
    },
    "non_linearity_coefficients": {
        "dim": [NL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "non linearity coefficients",
            "long_name": "non linearity coefficients",
            "units": "-",
        },
    },
    # "u_rel_random_non_linearity_coefficients": {"dim": [NL_DIM],
    #                           "dtype": np.uint32,
    #                           "attributes": {"standard_name": "random relative uncertainty on non linearity coefficients",
    #                                          "long_name": "random relative uncertainty on non linearity coefficients",
    #                                          "units": "-"},
    #                           "encoding": {'dtype': np.uint32, "scale_factor": 0.001, "offset": 0.0}},
    # "u_rel_systematic_indep_non_linearity_coefficients": {"dim": [NL_DIM],
    #                           "dtype": np.uint32,
    #                           "attributes": {"standard_name": "independent systematic relative uncertainty on non linearity coefficients",
    #                                          "long_name": "the systematic relative uncertainty component on non linearity coefficients that is not correlated between radiance and irradiance",
    #                                          "units": "-"},
    #                           "encoding": {'dtype': np.uint32, "scale_factor": 0.001, "offset": 0.0}},
    # "u_rel_systematic_corr_rad_irr_non_linearity_coefficients": {"dim": [NL_DIM],
    #                           "dtype": np.uint32,
    #                           "attributes": {"standard_name": "systematic relative uncertainty on non linearity coefficients (correlated radiance and irradiance)",
    #                                          "long_name": "the systematic relative uncertainty component on non linearity coefficients that is correlated between radiance and irradiance",
    #                                          "units": "-"},
    #                           "encoding": {'dtype': np.uint32, "scale_factor": 0.001, "offset": 0.0}},
    # "corr_random_non_linearity_coefficients":{"dim":[NL_DIM,NL_DIM],"dtype":np.int16,
    #                         "attributes":{
    #                             "standard_name":"correlation matrix of random error on non linearity coefficients",
    #                             "long_name":"Correlation matrix between wavelengths for the random errors on non linearity coefficients",
    #                             "units":"-"},
    #                         "encoding":{'dtype':np.int16,
    #                                     "scale_factor":0.001,"offset":0.0}},
    # "corr_systematic_indep_non_linearity_coefficients":{"dim":[NL_DIM,NL_DIM],
    #                                   "dtype":np.int16,"attributes":{
    #         "standard_name":"independent correlation matrix of systematic error on non linearity coefficients",
    #         "long_name":"Correlation matrix bewtween wavelengths for the systematic error component on non linearity coefficients that is not correlated between radiance and irradiance",
    #         "units":"-"},"encoding":{'dtype':np.int16,
    #                                                    "scale_factor":0.001,
    #                                                    "offset":0.0}},
    # "corr_systematic_corr_rad_irr_non_linearity_coefficients":{"dim":[NL_DIM,NL_DIM],
    #                                          "dtype":np.int16,"attributes":{
    #         "standard_name":"correlation matrix of systematic error on non linearity coefficients (correlated radiance and irradiance)",
    #         "long_name":"Correlation matrix bewtween wavelengths for the systematic error component on non linearity coefficients that is correlated between radiance and irradiance",
    #         "units":"-"},"encoding":{'dtype':np.int16,
    #                                                    "scale_factor":0.001,
    #                                                    "offset":0.0}},
}


# > L1A_RAD_VARIABLES - Radiance Variables required for L1 data products (except water L1B)
L1A_RAD_VARIABLES = {
    "radiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "radiance",
            "long_name": "upwelling radiance",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_radiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint16,
        "attributes": {
            "standard_name": "random relative uncertainty on radiance",
            "long_name": "random relative uncertainty on upwelling radiance",
            "units": "%",
            "err_corr": [
                {"dim": SCAN_DIM, "form": "random", "params": [], "units": []},
                {"dim": WL_DIM, "form": "random", "params": [], "units": []},
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_indep_radiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint16,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on radiance",
            "long_name": "the systematic relative uncertainty component on radiance that is not correlated with irradiance",
            "units": "%",
            "err_corr": [
                {"dim": SCAN_DIM, "form": "systematic", "params": [], "units": []},
                {
                    "dim": WL_DIM,
                    "form": "err_corr_matrix",
                    "params": ["corr_systematic_indep_radiance"],
                    "units": [],
                },
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_radiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint16,
        "attributes": {
            "standard_name": "systematic relative uncertainty on radiance, correlated with irradiance",
            "long_name": "the systematic relative uncertainty component on radiance that is correlated with irradiance",
            "units": "-",
            "err_corr": [
                {"dim": SCAN_DIM, "form": "systematic", "params": [], "units": []},
                {
                    "dim": WL_DIM,
                    "form": "err_corr_matrix",
                    "params": ["corr_systematic_corr_rad_irr_radiance"],
                    "units": [],
                },
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    # "corr_random_radiance": {"dim": [WL_DIM, WL_DIM],
    #                          "dtype": np.float32,
    #                          "attributes": {
    #                              "standard_name": "correlation matrix of random error on radiance",
    #                              "long_name": "Correlation matrix between wavelengths for the random errors on radiance",
    #                              "units": "-"},
    #                          "encoding": {'dtype': np.int8, "scale_factor": 0.01, "offset": 0.0}},
    "corr_systematic_indep_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radiance that is not correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on radiance, correlated with irradiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radiance that is correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
}

L1A_IRR_VARIABLES = {
    "irradiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "irradiance",
            "long_name": "downwelling irradiance",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_irradiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint16,
        "attributes": {
            "standard_name": "random relative uncertainty on irradiance",
            "long_name": "random relative uncertainty on downwelling irradiance",
            "units": "-",
            "err_corr": [
                {"dim": SCAN_DIM, "form": "random", "params": [], "units": []},
                {"dim": WL_DIM, "form": "random", "params": [], "units": []},
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_indep_irradiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint16,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on irradiance",
            "long_name": "the systematic relative uncertainty component on downwelling irradiance that is not correlated with radiance",
            "units": "-",
            "err_corr": [
                {"dim": SCAN_DIM, "form": "systematic", "params": [], "units": []},
                {
                    "dim": WL_DIM,
                    "form": "err_corr_matrix",
                    "params": ["corr_systematic_indep_irradiance"],
                    "units": [],
                },
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_irradiance": {
        "dim": [WL_DIM, SCAN_DIM],
        "dtype": np.uint16,
        "attributes": {
            "standard_name": "systematic relative uncertainty on irradiance, correlated with radiance",
            "long_name": "the systematic relative uncertainty component on downwelling irradiance that is correlated with radiance",
            "units": "-",
            "err_corr": [
                {"dim": SCAN_DIM, "form": "systematic", "params": [], "units": []},
                {
                    "dim": WL_DIM,
                    "form": "err_corr_matrix",
                    "params": ["corr_systematic_corr_rad_irr_irradiance"],
                    "units": [],
                },
            ],
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    # "corr_random_irradiance": {"dim": [WL_DIM, WL_DIM],
    #                            "dtype": np.uint16,
    #                            "attributes": {
    #                                "standard_name": "correlation matrix of random error on irradiance",
    #                                "long_name": "Correlation matrix between wavelengths for the random errors on irradiance",
    #                                "units": "-"},
    #                            "encoding": {'dtype': np.int8, "scale_factor": 0.01, "offset": 0.0}},
    "corr_systematic_indep_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on irradiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on irradiance that is not correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on irradiance, correlated with radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on irradiance that is correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
}

L1B_RAD_VARIABLES = {
    "radiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "radiance",
            "long_name": "upwelling radiance",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_radiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "random relative uncertainty on radiance",
            "long_name": "random relative uncertainty on upwelling radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_indep_radiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on radiance",
            "long_name": "the systematic relative uncertainty component on upwelling radiance that is not correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_radiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "systematic relative uncertainty on radiance, correlated with irradiance",
            "long_name": "the systematic relative uncertainty component on upwelling radiance that is correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of random error on radiance",
            "long_name": "Correlation matrix between wavelengths for the random errors on radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_indep_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radiance that is not correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on radiance, correlated with irradiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radianc that is correlated with irradiancee",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
}


L1B_IRR_VARIABLES = {
    "irradiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "irradiance",
            "long_name": "downwelling irradiance",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_irradiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "random relative uncertainty on irradiance",
            "long_name": "random relative uncertainty on downwelling irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_indep_irradiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on irradiance",
            "long_name": "the systematic relative uncertainty component on downwelling irradiance that is not correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_irradiance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "systematic relative uncertainty on irradiance, correlated with radiance",
            "long_name": "the systematic relative uncertainty component on downwelling irradiance that is correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of random error on irradiance",
            "long_name": "Correlation matrix between wavelengths for the random errors on irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_indep_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on irradiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on irradiance that is not correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on irradiance, correlated with radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on irradiance that is correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
}


W_L1C_VARIABLES = {
    "downwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "radiance",
            "long_name": "downwelling radiance",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_downwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "random relative uncertainty on radiance",
            "long_name": "random relative uncertainty on upwelling radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_indep_downwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on radiance",
            "long_name": "the systematic relative uncertainty component on upwelling radiance that is not correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_downwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "systematic relative uncertainty on radiance",
            "long_name": "the systematic relative uncertainty component on upwelling radiance that is correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_downwelling_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of random error on radiance",
            "long_name": "Correlation matrix between wavelengths for the random errors on radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_indep_downwelling_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radiance that is not correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_downwelling_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radiance that is correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "upwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "radiance",
            "long_name": "upwelling radiance",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_upwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "random relative uncertainty on radiance",
            "long_name": "random relative uncertainty on upwelling radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_indep_upwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on radiance",
            "long_name": "the systematic relative uncertainty component on upwelling radiance that is not correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_upwelling_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "systematic relative uncertainty on radiance",
            "long_name": "the systematic relative uncertainty component on upwelling radiance that is correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_upwelling_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of random error on radiance",
            "long_name": "Correlation matrix between wavelengths for the random errors on radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_indep_upwelling_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radiance that is not correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_upwelling_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on radiance that is correlated with irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "irradiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "irradiance",
            "long_name": "downwelling irradiance",
            "units": "mW m^-2 nm^-1 sr^-1",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_irradiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "random relative uncertainty on irradiance",
            "long_name": "random relative uncertainty on downwelling irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_indep_irradiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent systematic relative uncertainty on irradiance",
            "long_name": "the systematic relative uncertainty component on downwelling irradiance that is not correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_corr_rad_irr_irradiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "systematic relative uncertainty on irradiance, correlated with radiance",
            "long_name": "the systematic relative uncertainty component on downwelling irradiance that is correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of random error on irradiance",
            "long_name": "Correlation matrix between wavelengths for the random errors on irradiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_indep_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "independent correlation matrix of systematic error on irradiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on irradiance that is not correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_corr_rad_irr_irradiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "correlation matrix of systematic error on irradiance, correlated with radiance",
            "long_name": "Correlation matrix bewtween wavelengths for the systematic error component on irradiance that is correlated with radiance",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "rhof": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "fresnel_reflectance",
            "long_name": "Fraction of downwelling sky radiance reflected at the "
            "air-water interface",
            "reference": "SYSTEM_HEIGHT_DEPLOYEMENT",
            "units": "-",
            "preferred_symbol": "rhof",
        },
    },
    "fresnel_wind": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float16,
        "attributes": {
            "standard_name": "fresnel_wind",
            "long_name": "Surface wind speed used for the retrieval of the "
            "fraction of downwelling sky radiance reflected at "
            "the air-water interface",
            "reference": "SYSTEM_HEIGHT_DEPLOYEMENT",
            "units": "ms^-1",
        },
    },
    "fresnel_sza": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "fresnel_solar_zenith_angle",
            "long_name": "Solar zenith angle used for the retrieval of the "
            "fraction of downwelling sky radiance reflected at "
            "the air-water interface",
            "reference": "SYSTEM_HEIGHT_DEPLOYEMENT",
            "units": "degrees",
        },
    },
    "fresnel_raa": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "fresnel_relative_azimuth_angle",
            "long_name": "Relative azimuth angle from sun to sensor (0° when "
            "sun and sensor are aligned 180° when the sensor is "
            "looking into the sunglint) used for the retrieval "
            "of the fraction of downwelling sky radiance "
            "reflected at the air-water interface",
            "reference": "SYSTEM_HEIGHT_DEPLOYEMENT",
            "units": "degrees",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0055, "offset": 0.0},
    },
    "fresnel_vza": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "fresnel_sensor_zenith_angle",
            "long_name": "Sensor zenith angle used for the retrieval of the "
            "fraction of downwelling sky radiance reflected at "
            "the air-water interface",
            "reference": "",
            "units": "degrees",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0028, "offset": 0.0},
    },
}

# L_L2A_REFLECTANCE_VARIABLES - Reflectance variables required for L2A land data product
L_L2A_REFLECTANCE_VARIABLES = {
    "reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "reflectance",
            "long_name": "The surface called surface means the lower "
            "boundary of the atmosphere. "
            "Bidirectional_reflectance depends on"
            "the angles of incident and measured"
            "radiation.Reflectance is the ratio of the "
            "energy of the reflected to the incident "
            "radiation. A coordinate variable of "
            "radiation_wavelength or radiation_frequency "
            "can be used to specify the wavelength or "
            "frequency, respectively, of the radiation.",
            "units": "-",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_random_reflectance",
            "long_name": "Random reflectance relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_systematic_reflectance",
            "long_name": "Systematic reflectance " "relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_reflectance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_random_reflectance",
            "long_name": "Correlation matrix of random " "reflectance " "uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_reflectance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_systematic_reflectance",
            "long_name": "Correlation matrix of systematic "
            "reflectance "
            "uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
}

# W_L2A_REFLECTANCE_VARIABLES - Reflectance variables required for L2A water data product
W_L1C_REFLECTANCE_VARIABLES = {
    "reflectance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "surface_upwelling_radiance_per_unit_"
            "wavelength_in_air_reflected_by_water",
            "long_name": "The surface called 'surface' means the "
            "lower boundary of the atmosphere. "
            "Upwelling radiation is radiation from "
            "below. It does not mean 'net upward''. "
            "The sign convention is that 'upwelling' "
            "is positive upwards and 'downwelling' "
            "is positive downwards. Radiance is the "
            "radiative flux in a particular "
            "direction, per unit of solid angle. The "
            "direction towards which it is going must be"
            " specified, for instance with a coordinate "
            "of zenith_angle. ",
            "reference": "SYSTEM_HEIGHT_DEPLOYEMENT",
            "units": "-",
            "preferred_symbol": "rhow",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "water_leaving_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "water_leaving_radiance",
            "long_name": "water-leaving radiance"
            " of electromagnetic radiation "
            "(unspecified single wavelength)"
            " from the water body by "
            "cosine-collector radiometer",
            "reference": "",
            "units": "mW m^-2 nm^-1 sr^-1",
            "preferred_symbol": "lw",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "reflectance_nosc": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "water_leaving_reflectance_nosc",
            "long_name": "Reflectance of the water column at the "
            "surface without correction for the NIR "
            "similarity spectrum "
            "(see Ruddick et al., 2006)",
            "units": "-",
            "preferred_symbol": "rhow_nosc",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_water_leaving_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_random_normalized_water_leaving_radiance",
            "long_name": "Random normalized water leaving radiance "
            "relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_water_leaving_radiance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_systematic_normalized_water_leaving_radiance",
            "long_name": "Systematic normalized water leaving radiance "
            "relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_water_leaving_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_random_normalized_water_leaving_radiance",
            "long_name": "Correlation matrix of random normalized water "
            "leaving radiance uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_water_leaving_radiance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_systematic_normalized_water_leaving_radiance",
            "long_name": "Correlation matrix of systematic normalized water "
            "leaving radiance uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "u_rel_random_reflectance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_random_water_leaving_reflectance",
            "long_name": "Random water leaving reflectance " "relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_reflectance": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_systematic_water_leaving_reflectance",
            "long_name": "Systematic water leaving reflectance relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_reflectance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_random_water_leaving_reflectance",
            "long_name": "Correlation matrix of random water leaving "
            "reflectance uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_reflectance": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_systematic_water_leaving_reflectance",
            "long_name": "Correlation matrix of systematic water leaving "
            "reflectance uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "u_rel_random_reflectance_nosc": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_random_water_leaving_reflectance_nosc",
            "long_name": "Random water leaving reflectance not corrected "
            "for NIR similarity spectrum relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_reflectance_nosc": {
        "dim": [WL_DIM, Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_systematic_water_leaving_reflectance_nosc",
            "long_name": "Systematic water leaving reflectance not "
            "corrected for NIR similarity spectrum "
            "relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_reflectance_nosc": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_random_water_leaving_reflectance_nosc",
            "long_name": "Correlation matrix of random water leaving "
            "reflectance not corrected for NIR similarity "
            "spectrum uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_reflectance_nosc": {
        "dim": [WL_DIM, WL_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_systematic_water_leaving_reflectance_nosc",
            "long_name": "Correlation matrix of systematic water "
            "leaving reflectance not corrected for NIR "
            "similarity spectrum uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "epsilon": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "simil_epsilon",
            "long_name": "Similarity spectrum ratio at two wavelengths see Ruddick et al. (2016)",
            "reference": "",
            "units": "-",
        },
    },
    "u_rel_random_epsilon": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_random_epsilon",
            "long_name": "Random epsilon " "relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_epsilon": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_systematic_epsilon",
            "long_name": "Systematic epsilon relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "rhof": {
        "dim": [Lu_SCAN_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "fresnel_reflectance",
            "long_name": "Fraction of downwelling sky radiance reflected at the "
            "air-water interface",
            "reference": "SYSTEM_HEIGHT_DEPLOYEMENT",
            "units": "-",
            "preferred_symbol": "rhof",
        },
    },
}
W_L2A_REFLECTANCE_VARIABLES = deepcopy(W_L1C_REFLECTANCE_VARIABLES)

for variable in W_L2A_REFLECTANCE_VARIABLES.keys():
    W_L2A_REFLECTANCE_VARIABLES[variable]["dim"] = [
        d if d != Lu_SCAN_DIM else SERIES_DIM
        for d in W_L2A_REFLECTANCE_VARIABLES[variable]["dim"]
    ]

# L_L2B_REFLECTANCE_VARIABLES - Reflectance variables required for L2A land data product
L_L2B_REFLECTANCE_VARIABLES = {
    "reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "reflectance",
            "long_name": "The surface called surface means the lower "
            "boundary of the atmosphere. "
            "Bidirectional_reflectance depends on"
            "the angles of incident and measured"
            "radiation.Reflectance is the ratio of the "
            "energy of the reflected to the incident "
            "radiation. A coordinate variable of "
            "radiation_wavelength or radiation_frequency "
            "can be used to specify the wavelength or "
            "frequency, respectively, of the radiation.",
            "units": "-",
            "unc_comps": [
                "u_rel_random_radiance",
                "u_rel_systematic_indep_radiance",
                "u_rel_systematic_corr_rad_irr_radiance",
            ],
        },
    },
    "u_rel_random_reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_random_reflectance",
            "long_name": "Random reflectance relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "u_rel_systematic_reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "u_rel_systematic_reflectance",
            "long_name": "Systematic reflectance " "relative uncertainty",
            "units": "%",
        },
        "encoding": {"dtype": np.uint16, "scale_factor": 0.0001, "offset": 0.0},
    },
    "corr_random_reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_random_reflectance",
            "long_name": "Correlation matrix of random " "reflectance " "uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
    "corr_systematic_reflectance": {
        "dim": [WL_DIM, SERIES_DIM],
        "dtype": np.float32,
        "attributes": {
            "standard_name": "corr_systematic_reflectance",
            "long_name": "Correlation matrix of systematic "
            "reflectance "
            "uncertainty",
            "units": "-",
        },
        "encoding": {"dtype": np.int8, "scale_factor": 0.01, "offset": 0.0},
    },
}

# W_L2B_REFLECTANCE_VARIABLES - Reflectance variables required for L2A land data product
W_L2B_REFLECTANCE_VARIABLES = {}

# C_QUALITY_SCAN = {"quality_flag": {"dim": [SCAN_DIM],
#                                    "dtype": "flag",
#                                    "attributes": {"standard_name": "quality_flag",
#                                                   "long_name": "A variable with the standard name of quality_"
#                                                                "flag contains an indication of assessed "
#                                                                "quality information of another data variable."
#                                                                " The linkage between the data variable and the"
#                                                                " variable or variables with the standard_name"
#                                                                " of quality_flag is achieved using the "
#                                                                "ancillary_variables attribute.",
#                                                   "flag_meanings": FLAG_COMMON},
#                                    "encoding": {}}}
#
# L_QUALITY = {"quality_flag": {"dim": [SERIES_DIM],
#                               "dtype": "flag",
#                               "attributes": {"standard_name": "quality_flag",
#                                              "long_name": "A variable with the standard name of quality_"
#                                                           "flag contains an indication of assessed "
#                                                           "quality information of another data variable."
#                                                           " The linkage between the data variable and the"
#                                                           " variable or variables with the standard_name"
#                                                           " of quality_flag is achieved using the "
#                                                           "ancillary_variables attribute.",
#                                              "flag_meanings": FLAG_COMMON + FLAG_LAND},
#                               "encoding": {}}}
#
# W_QUALITY_SERIES = {"quality_flag": {"dim": [SERIES_DIM],
#                                      "dtype": "flag",
#                                      "attributes": {"standard_name": "quality_flag",
#                                                     "long_name": "A variable with the standard name of quality_"
#                                                                  "flag contains an indication of assessed "
#                                                                  "quality information of another data variable."
#                                                                  " The linkage between the data variable and the"
#                                                                  " variable or variables with the standard_name"
#                                                                  " of quality_flag is achieved using the "
#                                                                  "ancillary_variables attribute.",
#                                                     "flag_meanings": FLAG_COMMON + FLAG_WATER},
#                                      "encoding": {}}}
#
# W_QUALITY_SCAN = deepcopy(W_QUALITY_SERIES)
# for variable in W_QUALITY_SCAN.keys():
#     W_QUALITY_SCAN[variable]["dim"] = [d if d != SERIES_DIM else Lu_SCAN_DIM
#                                        for d in W_QUALITY_SCAN[variable]["dim"]]
# File format variable defs
# -------------------------

VARIABLES_DICT_DEFS: Any = {
    "L0_RAD": {**COMMON_VARIABLES_SCAN, **L0_RAD_VARIABLES},
    "L0_IRR": {**COMMON_VARIABLES_SCAN, **L0_IRR_VARIABLES},
    "L0_BLA": {**COMMON_VARIABLES_SCAN, **L0_BLA_VARIABLES},
    "CAL": {**CAL_VARIABLES},
    "L_L1A_RAD": {**COMMON_VARIABLES_SCAN, **L1A_RAD_VARIABLES},
    "W_L1A_RAD": {**COMMON_VARIABLES_SCAN, **L1A_RAD_VARIABLES},
    "L_L1A_IRR": {**COMMON_VARIABLES_SCAN, **L1A_IRR_VARIABLES},
    "W_L1A_IRR": {**COMMON_VARIABLES_SCAN, **L1A_IRR_VARIABLES},
    "L_L1B_RAD": {**COMMON_VARIABLES_SERIES, **L1B_RAD_VARIABLES},
    "L_L1B_IRR": {**COMMON_VARIABLES_SERIES, **L1B_IRR_VARIABLES},
    "W_L1B_RAD": {**COMMON_VARIABLES_SERIES, **L1B_RAD_VARIABLES},
    "W_L1B_IRR": {**COMMON_VARIABLES_SERIES, **L1B_IRR_VARIABLES},
    "L_L1C": {**COMMON_VARIABLES_SERIES, **L1B_RAD_VARIABLES, **L1B_IRR_VARIABLES},
    "W_L1C": {
        **COMMON_VARIABLES_Lu_SCAN,
        **W_L1C_VARIABLES,
        **W_L1C_REFLECTANCE_VARIABLES,
    },
    "L_L2A": {**COMMON_VARIABLES_SERIES, **L_L2A_REFLECTANCE_VARIABLES},
    "W_L2A": {**COMMON_VARIABLES_SERIES, **W_L2A_REFLECTANCE_VARIABLES},
    "L_L2B": {**COMMON_VARIABLES_SERIES, **L_L2B_REFLECTANCE_VARIABLES},
}



"""Use Monte Carlo to propagate uncertainties"""

import time
import warnings
from multiprocessing import Pool

import comet_maths as cm
import numpy as np

import punpy.utilities.utilities as util

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


class MCPropagation:
    def __init__(self, steps, parallel_cores=0, dtype=None, verbose=False, MCdimlast=True):
        """
        Initialise MC Propagator

        :param steps: number of MC iterations
        :type steps: int
        :param parallel_cores: number of CPU to be used in parallel processing
        :type parallel_cores: int
        :param dtype: numpy dtype for output variables
        :type dtype: numpy dtype
        :param verbose: bool to set if logging info should be printed
        :type verbose: bool
        :param MCdimlast: bool to set whether the MC dimension should be moved to the last dimension when running through the measurment function (when parallel_cores==0). This can be useful for broadcasting within the measurement function. defaults to False
        :type MCdimlast: bool
        """

        self.MCsteps = steps
        self.parallel_cores = parallel_cores
        self.dtype = dtype
        if parallel_cores > 1:
            self.pool = Pool(self.parallel_cores)
        self.verbose = verbose
        self.starttime = time.time()
        self.MCdimlast = MCdimlast

    def propagate_random(
        self,
        func,
        x,
        u_x,
        corr_x=None,
        param_fixed=None,
        corr_between=None,
        samples=None,
        return_corr=False,
        return_samples=False,
        repeat_dims=-99,
        corr_dims=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
        refyvar=0,
        pdf_shape="gaussian",
        pdf_params=None,
            allow_some_nans=True,
    ):
        """
        Propagate random uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Random uncertainties arise when there is no correlation between repeated measurements.
        It is possible (though rare) that there is a correlation in one of the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of random uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param samples: allows to provide a Monte Carlo sample previously generated. This sample of input quantities will be used instead of generating one from the uncertainties and error-correlation. Defaults to None
        :type samples: list[array], optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_dims: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int, optional
        :param pdf_shape: string identifier of the probability density function shape, defaults to gaussian
        :type pdf_shape: str, optional
        :param pdf_params: dictionaries defining optional additional parameters that define the probability density function, Defaults to None (gaussian does not require additional parameters)
        :type pdf_params: dict, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        if corr_x is None:
            corr_x = ["rand"] * len(x)
        for i in range(len(x)):
            if corr_x[i] is None:
                corr_x[i] = "rand"

        return self.propagate_standard(
            func,
            x,
            u_x,
            corr_x,
            param_fixed=param_fixed,
            corr_between=corr_between,
            samples=samples,
            return_corr=return_corr,
            return_samples=return_samples,
            repeat_dims=repeat_dims,
            corr_dims=corr_dims,
            fixed_corr_var=fixed_corr_var,
            output_vars=output_vars,
            PD_corr=PD_corr,
            refyvar=refyvar,
            pdf_shape=pdf_shape,
            pdf_params=pdf_params,
            allow_some_nans=allow_some_nans
        )

    def propagate_systematic(
        self,
        func,
        x,
        u_x,
        corr_x=None,
        param_fixed=None,
        corr_between=None,
        samples=None,
        return_corr=False,
        return_samples=False,
        repeat_dims=-99,
        corr_dims=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
        refyvar=0,
        pdf_shape="gaussian",
        pdf_params=None,
            allow_some_nans=True,
    ):
        """
        Propagate systematic uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Systematic uncertainties arise when there is full correlation between repeated measurements.
        There is a often also a correlation between measurements along the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param samples: allows to provide a Monte Carlo sample previously generated. This sample of input quantities will be used instead of generating one from the uncertainties and error-correlation. Defaults to None
        :type samples: list[array], optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_dims: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int, optional
        :param pdf_shape: string identifier of the probability density function shape, defaults to gaussian
        :type pdf_shape: str, optional
        :param pdf_params: dictionaries defining optional additional parameters that define the probability density function, Defaults to None (gaussian does not require additional parameters)
        :type pdf_params: dict, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        if corr_x is None:
            corr_x = ["syst"] * len(x)
        for i in range(len(x)):
            if corr_x[i] is None:
                corr_x[i] = "syst"

        return self.propagate_standard(
            func,
            x,
            u_x,
            corr_x,
            param_fixed=param_fixed,
            corr_between=corr_between,
            samples=samples,
            return_corr=return_corr,
            return_samples=return_samples,
            repeat_dims=repeat_dims,
            corr_dims=corr_dims,
            fixed_corr_var=fixed_corr_var,
            output_vars=output_vars,
            PD_corr=PD_corr,
            refyvar=refyvar,
            pdf_shape=pdf_shape,
            pdf_params=pdf_params,
            allow_some_nans=allow_some_nans
        )

    def propagate_cov(
        self,
        func,
        x,
        cov_x,
        param_fixed=None,
        corr_between=None,
        samples=None,
        return_corr=True,
        return_samples=False,
        repeat_dims=-99,
        corr_dims=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
        refyvar=0,
        pdf_shape="gaussian",
        pdf_params=None,
            allow_some_nans=True,
    ):
        """
        Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        The covariance matrix can represent the full covariance matrix between all measurements in all dimensions.
        Alternatively if there are repeated measurements specified in repeat_dims, the covariance matrix is given
        for the covariance along the dimension that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param cov_x: list of covariance matrices on input quantities (usually numpy arrays). In case the input quantity is an array of shape (m,o), the covariance matrix  is typically given as an array of shape (m*o,m*o).
        :type cov_x: list[array]
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param samples: allows to provide a Monte Carlo sample previously generated. This sample of input quantities will be used instead of generating one from the uncertainties and error-correlation. Defaults to None
        :type samples: list[array], optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_dims: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int, optional
        :param pdf_shape: string identifier of the probability density function shape, defaults to gaussian
        :type pdf_shape: str, optional
        :param pdf_params: dictionaries defining optional additional parameters that define the probability density function, Defaults to None (gaussian does not require additional parameters)
        :type pdf_params: dict, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        x = [np.atleast_1d(x[i]) for i in range(len(x))]
        cov_x = [np.atleast_2d(cov_x[i]) for i in range(len(x))]
        u_x = [
            cm.uncertainty_from_covariance(cov_x[i]).reshape(x[i].shape)
            for i in range(len(x))
        ]
        corr_x = [cm.correlation_from_covariance(cov_x[i]) for i in range(len(x))]

        return self.propagate_standard(
            func,
            x,
            u_x,
            corr_x,
            param_fixed=param_fixed,
            corr_between=corr_between,
            samples=samples,
            return_corr=return_corr,
            return_samples=return_samples,
            repeat_dims=repeat_dims,
            corr_dims=corr_dims,
            fixed_corr_var=fixed_corr_var,
            output_vars=output_vars,
            PD_corr=PD_corr,
            refyvar=refyvar,
            pdf_shape=pdf_shape,
            pdf_params=pdf_params,
            allow_some_nans=allow_some_nans
        )

    def propagate_standard(
        self,
        func,
        x,
        u_x,
        corr_x,
        param_fixed=None,
        corr_between=None,
        samples=None,
        return_corr=False,
        return_samples=False,
        repeat_dims=-99,
        corr_dims=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
        refyvar=0,
        pdf_shape="gaussian",
        pdf_params=None,
        allow_some_nans=True,
    ):
        """
        Propagate uncertainties through measurement function with n input quantities. Correlations must be specified in corr_x.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Systematic uncertainties arise when there is full correlation between repeated measurements.
        There is a often also a correlation between measurements along the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis. Can be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array]
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param samples: allows to provide a Monte Carlo sample previously generated. This sample of input quantities will be used instead of generating one from the uncertainties and error-correlation. Defaults to None
        :type samples: list[array], optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_dims: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int, optional
        :param pdf_shape: string identifier of the probability density function shape, defaults to gaussian
        :type pdf_shape: str, optional
        :param pdf_params: dictionaries defining optional additional parameters that define the probability density function, Defaults to None (gaussian does not require additional parameters)
        :type pdf_params: dict, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        if self.verbose:
            print(
                "starting propagation (%s s since creation of prop object)"
                % (time.time() - self.starttime)
            )
        (
            yshapes,
            x,
            u_x,
            corr_x,
            n_repeats,
            repeat_shape,
            repeat_dims,
            corr_dims,
            fixed_corr,
        ) = self.perform_checks(
            func,
            x,
            u_x,
            corr_x,
            repeat_dims,
            corr_dims,
            output_vars,
            fixed_corr_var,
            param_fixed,
            refyvar,
        )

        if n_repeats > 0:
            xb, u_xb = util.select_repeated_x(
                x, u_x, param_fixed, 0, repeat_dims, repeat_shape
            )
            out_0 = self.propagate_standard(
                func,
                xb,
                u_xb,
                corr_x,
                param_fixed,
                corr_between,
                samples,
                return_corr,
                return_samples,
                -99,
                corr_dims=corr_dims,
                fixed_corr_var=fixed_corr_var,
                output_vars=output_vars,
                PD_corr=False,
                pdf_shape=pdf_shape,
                pdf_params=pdf_params,
            )

            outs = self.make_new_outs(
                out_0,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
                refyvar,
            )

            outs = self.add_repeated_outs(
                outs,
                0,
                out_0,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
                refyvar,
            )

            for i in range(1, n_repeats):
                xb, u_xb = util.select_repeated_x(
                    x, u_x, param_fixed, i, repeat_dims, repeat_shape
                )
                print("here",i,u_xb)
                out_i = self.propagate_standard(
                    func,
                    xb,
                    u_xb,
                    corr_x,
                    param_fixed,
                    corr_between,
                    samples,
                    return_corr,
                    return_samples,
                    -99,
                    corr_dims=corr_dims,
                    fixed_corr_var=fixed_corr_var,
                    output_vars=output_vars,
                    PD_corr=False,
                    pdf_shape=pdf_shape,
                    pdf_params=pdf_params,
                )

                outs = self.add_repeated_outs(
                    outs,
                    i,
                    out_i,
                    yshapes,
                    len(x),
                    n_repeats,
                    repeat_shape,
                    repeat_dims,
                    return_corr,
                    return_samples,
                    output_vars,
                    refyvar,
                )
                if self.verbose:
                    print(
                        "repeated measurement %s out of %s processed (%s s since creation of prop object)"
                        % (i, n_repeats, time.time() - self.starttime)
                    )

            outs = self.finish_repeated_outs(
                outs,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
                refyvar,
            )

            return outs

        else:
            if samples is not None:
                MC_x = samples
            else:
                MC_x = self.generate_MC_sample(
                    x,
                    u_x,
                    corr_x,
                    corr_between=corr_between,
                    pdf_shape=pdf_shape,
                    pdf_params=pdf_params,
                )

            MC_y = self.run_samples(func, MC_x, output_vars=output_vars,allow_some_nans=allow_some_nans)

            return self.process_samples(
                MC_x,
                MC_y,
                return_corr,
                return_samples,
                yshapes,
                corr_dims,
                fixed_corr,
                PD_corr,
                output_vars,
            )

    def generate_MC_sample(
        self,
        x,
        u_x,
        corr_x,
        corr_between=None,
        pdf_shape="gaussian",
        pdf_params=None,
        comp_list=False,
    ):
        """
        function to generate MC sample for input quantities

        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis. Can be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array]
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param pdf_shape: string identifier of the probability density function shape, defaults to gaussian
        :type pdf_shape: str, optional
        :param pdf_params: dictionaries defining optional additional parameters that define the probability density function, Defaults to None (gaussian does not require additional parameters)
        :type pdf_params: dict, optional
        :param comp_list: boolean to define whether u_x and corr_x are given as a list or individual uncertainty components. Defaults to False, in chich case a single combined uncertainty component is given per input quantity.
        :type comp_list: bool, optional
        :return: MC sample for input quantities
        :rtype: list[array]
        """
        MC_data = np.empty(len(x), dtype=np.ndarray)

        for i in range(len(x)):
            MC_data[i] = cm.generate_sample(
                self.MCsteps,
                x,
                u_x,
                corr_x,
                i,
                pdf_shape=pdf_shape,
                pdf_params=pdf_params,
                comp_list=comp_list,
            )
            print(MC_data[i],u_x[i],corr_x[i])

        if corr_between is not None:
            MC_data = cm.correlate_sample_corr(MC_data, corr_between)
        if self.verbose:
            print(
                "samples generated (%s s since creation of prop object)"
                % (time.time() - self.starttime)
            )
        return MC_data

    def generate_MC_sample_cov(
        self, x, cov_x, corr_between=None, pdf_shape="gaussian", pdf_params=None
    ):
        """
        function to generate MC sample for input quantities from covariance matrix

        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis. Can be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array]
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param pdf_shape: string identifier of the probability density function shape, defaults to gaussian
        :type pdf_shape: str, optional
        :param pdf_params: dictionaries defining optional additional parameters that define the probability density function, Defaults to None (gaussian does not require additional parameters)
        :type pdf_params: dict, optional
        :return: MC sample for input quantities
        :rtype:
        """

        MC_data = np.empty(len(x), dtype=np.ndarray)
        for i in range(len(x)):
            if not hasattr(x[i], "__len__"):
                MC_data[i] = cm.generate_sample_systematic(
                    self.MCsteps,
                    x[i],
                    cov_x[i],
                    pdf_shape=pdf_shape,
                    pdf_params=pdf_params,
                )
            else:
                MC_data[i] = cm.generate_sample_cov(
                    self.MCsteps,
                    x[i].ravel(),
                    cov_x[i],
                    pdf_shape=pdf_shape,
                    pdf_params=pdf_params,
                ).reshape(x[i].shape + (self.MCsteps,))

        if corr_between is not None:
            MC_data = cm.correlate_sample_corr(MC_data, corr_between)

        return MC_data

    def propagate_cov_flattened(
        self,
        func,
        x,
        cov_x,
        param_fixed=None,
        corr_between=None,
        samples=None,
        return_corr=True,
        return_samples=False,
        repeat_dims=-99,
        corr_dims=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
        refyvar=0,
        pdf_shape="gaussian",
        pdf_params=None,
    ):
        """
        Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        The covariance matrix can represent the full covariance matrix between all measurements in all dimensions.
        Alternatively if there are repeated measurements specified in repeat_dims, the covariance matrix is given
        for the covariance along the dimension that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param cov_x: list of covariance matrices on input quantities (usually numpy arrays). In case the input quantity is an array of shape (m,o), the covariance matrix needs to be given as an array of shape (m*o,m*o).
        :type cov_x: list[array]
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param samples: allows to provide a Monte Carlo sample previously generated. This sample of input quantities will be used instead of generating one from the uncertainties and error-correlation. Defaults to None
        :type samples: list[array], optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_dims: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int, optional
        :param pdf_shape: string identifier of the probability density function shape, defaults to gaussian
        :type pdf_shape: str, optional
        :param pdf_params: dictionaries defining optional additional parameters that define the probability density function, Defaults to None (gaussian does not require additional parameters)
        :type pdf_params: dict, optional
        :return: uncertainties on measurand
        :rtype: array
        """

        (
            yshapes,
            x,
            u_x,
            corr_x,
            n_repeats,
            repeat_shape,
            repeat_dims,
            corr_dims,
            fixed_corr,
        ) = self.perform_checks(
            func,
            x,
            cov_x,
            None,
            repeat_dims,
            corr_dims,
            output_vars,
            fixed_corr_var,
            param_fixed,
            refyvar,
        )

        if n_repeats > 0:
            xb, _ = util.select_repeated_x(
                x, x, param_fixed, 0, repeat_dims, repeat_shape
            )
            out_0 = self.propagate_cov(
                func,
                xb,
                cov_x,
                param_fixed,
                corr_between,
                return_corr,
                return_samples,
                -99,
                corr_dims=corr_dims,
                output_vars=output_vars,
                PD_corr=False,
                pdf_shape=pdf_shape,
                pdf_params=pdf_params,
            )

            outs = self.make_new_outs(
                out_0,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
                refyvar,
            )

            outs = self.add_repeated_outs(
                outs,
                0,
                out_0,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
                refyvar,
            )

            for i in range(1, n_repeats):
                xb, _ = util.select_repeated_x(
                    x, x, param_fixed, i, repeat_dims, repeat_shape
                )

                out_i = self.propagate_cov(
                    func,
                    xb,
                    cov_x,
                    param_fixed,
                    corr_between,
                    return_corr,
                    return_samples,
                    -99,
                    corr_dims=corr_dims,
                    output_vars=output_vars,
                    PD_corr=False,
                    pdf_shape=pdf_shape,
                    pdf_params=pdf_params,
                )

                outs = self.add_repeated_outs(
                    outs,
                    i,
                    out_i,
                    yshapes,
                    len(x),
                    n_repeats,
                    repeat_shape,
                    repeat_dims,
                    return_corr,
                    return_samples,
                    output_vars,
                    refyvar,
                )

            outs = self.finish_repeated_outs(
                outs,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
                refyvar,
            )

            return outs

        else:
            if samples is not None:
                MC_data = samples
            else:
                MC_data = self.generate_MC_sample_cov(
                    x,
                    cov_x,
                    corr_between=corr_between,
                    pdf_shape=pdf_shape,
                    pdf_params=pdf_params,
                )

        return self.process_samples(
            func,
            MC_data,
            return_corr,
            return_samples,
            yshapes,
            corr_dims,
            fixed_corr,
            PD_corr,
            output_vars,
        )

    def perform_checks(
        self,
        func,
        x,
        u_x,
        corr_x,
        repeat_dims,
        corr_dims,
        output_vars,
        fixed_corr_var,
        param_fixed,
        refyvar,
    ):
        """
        Perform checks on the input parameters and set up the appropriate keywords for further processing

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of uncertainties/covariances on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_dims: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int
        :return: yshape,u_x,repeat_axis,repeat_dims,corr_dims,fixed_corr
        :rtype: tuple, list[array], int, int, int, array
        """

        # Set up repeat_axis and repeat_dims for proper use in recursive function.
        if isinstance(repeat_dims, int):
            repeat_dims = [repeat_dims]

        # find the shape
        if output_vars == 1:
            yshape = np.array(func(*x)).shape
            yshapes = [np.array(func(*x)).shape]
        else:
            yshape = np.array(func(*x)[refyvar]).shape
            yshapes = [np.array(func(*x)[i]).shape for i in range(output_vars)]

        shapewarning = False
        for i in range(len(x)):
            try:
                if not (np.all(np.isfinite(u_x[i]))):
                    warnings.warn(
                        "One of your uncertainties has nans (of inf). This can cause issues in the uncertainty propagation. Please remove or replace your nans."
                    )

                if not (np.all(np.isfinite(corr_x[i]))):
                    warnings.warn(
                        "One of your error correlation matrices has nans (of inf). This can cause issues in the uncertainty propagation. Please remove or replace your nans."
                    )
            except:
                pass

            if hasattr(x[i], "__shape__"):
                if param_fixed is not None:
                    if (
                        x[i].shape != yshape
                        and self.parallel_cores == 0
                        and not param_fixed[i]
                    ):
                        shapewarning = True
                else:
                    if x[i].shape != yshape and self.parallel_cores == 0:
                        shapewarning = True
            else:
                if self.dtype is not None:
                    x[i] = np.array(x[i], dtype=self.dtype)
                if self.parallel_cores == 0 and not hasattr(x[i], "__len__"):
                    shapewarning = True

        if shapewarning:
            warnings.warn(
                "It looks like one of your input quantities is not an array or does not "
                "have the same shape as the measurand. This is not a problem, but means "
                "you likely cannot use array operations in your measurement function. "
                "You might need to set parallel_cores to 1 or higher when creating "
                "your MCPropagation object."
            )

        # Check for which input quantities there is no uncertainty,
        # replacing Nones with zeros where necessary.
        # Count the number of non-zero uncertainties. If this number is one, the
        # correlation matrix for the measurand will be the same as for this input qty.

        count = 0
        for i in range(len(x)):
            if u_x[i] is None:
                if hasattr(x[i], "__len__"):
                    u_x[i] = np.zeros(x[i].shape)
                else:
                    u_x[i] = 0.0

            if u_x is not None:
                if u_x[i] is not None:
                    if self.dtype is not None:
                        u_x[i] = np.array(u_x[i], dtype=self.dtype)

            if np.sum(u_x[i]) != 0 and fixed_corr_var == True:
                count += 1
                var = i

            if corr_x is not None:
                if corr_x[i] is not None:
                    if not (isinstance(corr_x[i], str) or isinstance(corr_x[i], dict)):
                        if np.any(corr_x[i] > 1.000001):
                            raise ValueError(
                                "punpy.mc_propagation: One of the provided correlation matrices "
                                "has elements >1."
                            )

        if count == 1:
            fixed_corr_var = var
        else:
            fixed_corr_var = -99

        if fixed_corr_var >= 0 and corr_x is not None:
            if corr_x[fixed_corr_var] == "rand":
                fixed_corr = np.eye(len(u_x[fixed_corr_var]))
            elif corr_x[fixed_corr_var] == "syst":
                fixed_corr = np.ones(
                    (len(u_x[fixed_corr_var]), len(u_x[fixed_corr_var]))
                )
            else:
                fixed_corr = corr_x[fixed_corr_var]

        else:
            fixed_corr = None

        if len(repeat_dims) == 1:
            if repeat_dims[0] >= 0:
                n_repeats = yshape[repeat_dims[0]]
                if corr_dims > repeat_dims[0]:
                    corr_dims -= 1
                elif corr_dims == repeat_dims[0]:
                    raise ValueError(
                        "punpy.mc_propagation: corr_dims and repeat_axis keywords should not be the same."
                    )
            else:
                n_repeats = 0
            repeat_shape = (n_repeats,)  # repeat_dims = -99
        else:
            repeat_dims = -np.sort(-np.array(repeat_dims))
            repeat_shape = tuple([yshape[repeat_dim] for repeat_dim in repeat_dims])
            n_repeats = np.prod(repeat_shape)
            for repeat_dim in repeat_dims:
                if corr_dims > repeat_dim:
                    corr_dims -= 1
                elif corr_dims == repeat_dim:
                    raise ValueError(
                        "punpy.mc_propagation: corr_dims and repeat_axis keywords should not be the same."
                    )

        return (
            yshapes,
            x,
            u_x,
            corr_x,
            n_repeats,
            repeat_shape,
            repeat_dims,
            corr_dims,
            fixed_corr,
        )

    def make_new_outs(
        self,
        out_0,
        yshapes,
        lenx,
        n_repeats,
        repeat_shape,
        repeat_dims,
        return_corr,
        return_samples,
        output_vars,
        refyvar,
    ):
        """
        Prepare arrays for storing the results from the individual repeated measurements.

        :param out_0: output array of the first repeated measurement
        :type out_0: array
        :type yshapes: tuple
        :param lenx: number of input quantities
        :type lenx: int
        :param n_repeats: number of repeated measurements
        :type n_repeats: int
        :param repeat_shape: shape along which the measurements are repeated
        :type repeat_shape: tuple
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian, defaults to False
        :type return_Jacobian: bool, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int
        :return: array of outputs of the repeated measurements
        :rtype: array
        """
        if not return_corr and output_vars == 1 and not return_samples:
            outs = np.empty((n_repeats,) + out_0.shape, dtype=self.dtype)

        else:
            outs = np.empty(len(out_0), dtype=object)

            if output_vars == 1:
                outs[0] = np.empty((n_repeats,) + out_0[0].shape, dtype=self.dtype)

            elif output_vars > 1 and not return_corr and not return_samples:
                outs[0] = np.empty(
                    (
                        output_vars,
                        n_repeats,
                    )
                    + out_0[0].shape,
                    dtype=self.dtype,
                )

            # elif (out_0[0][0].ndim) > 1:
            #     outs[0] = np.empty((output_vars,len(out_0[0][0]),n_repeats,)+out_0[0][0].shape)

            else:
                outs[0] = np.empty(
                    (
                        output_vars,
                        n_repeats,
                    )
                    + out_0[0][0].shape,
                    dtype=self.dtype,
                )

            extra_index = 0
            if return_corr:
                if output_vars == 1:
                    outs[1] = np.zeros_like(out_0[1], dtype=self.dtype)
                    extra_index += 1

                else:
                    outs[1] = np.empty((output_vars,), dtype=object)
                    for ii in range(output_vars):
                        outs[1][ii] = np.zeros_like(out_0[1][ii], dtype=self.dtype)

                    extra_index += 1
                    if all(
                        [yshapes[i] == yshapes[refyvar] for i in range(len(yshapes))]
                    ):
                        corr_out = np.zeros_like(
                            out_0[1 + extra_index], dtype=self.dtype
                        )
                    else:
                        corr_out = None
                    outs[1 + extra_index] = corr_out
                    extra_index += 1

            if return_samples:
                outs[1 + extra_index] = np.empty(lenx, dtype=object)
                outs[2 + extra_index] = np.empty(lenx, dtype=object)
        return outs

    def add_repeated_outs(
        self,
        outs,
        i,
        out_i,
        yshapes,
        lenx,
        n_repeats,
        repeat_shape,
        repeat_dims,
        return_corr,
        return_samples,
        output_vars,
        refyvar,
    ):
        """
        Add the results for a single repeated measurement to the combined array

        :param outs: array with outputs of the repeated measurements
        :type outs: list[array]
        :param i: index of the individual measurements
        :type i: int
        :param out_i: array with outputs of the individual measurements
        :type out_i: list[array]
        :param yshapes: shape of the measurand
        :type yshapes: tuple
        :param lenx: number of input quantities
        :type lenx: int
        :param n_repeats: number of repeated measurements
        :type n_repeats: int
        :param repeat_shape: shape along which the measurements are repeated
        :type repeat_shape: tuple
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian, defaults to False
        :type return_Jacobian: bool, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int
        :return: combined outputs
        :rtype: list[array]
        """

        if not return_corr and output_vars == 1 and not return_samples:
            outs[i] = out_i

        elif output_vars == 1:
            outs[0][i] = out_i[0]

        elif output_vars > 1 and not return_corr and not return_samples:
            outs[0]
            for ii in range(output_vars):
                outs[0][ii][i] = out_i[ii]

        # elif (out_i[0][0].ndim) > 1:
        #     outs[0] = np.empty(output_vars,dtype=object)
        #     for ii in range(output_vars):
        #         for iii in range(len(out_i[0][ii])):
        #             outs[0][ii][iii][i] = out_i[0][ii][iii]

        else:
            for ii in range(output_vars):
                outs[0][ii][i] = out_i[0][ii]

        extra_index = 0
        if return_corr:
            if output_vars == 1:
                corr = out_i[1]
                if np.isnan(corr).any():
                    warnings.warn(
                        "one of the measurements along the repeat_dim has nans"
                    )

                else:
                    outs[1] += corr
                extra_index += 1

            else:
                for ii in range(output_vars):
                    corr = out_i[1][ii]
                    if np.isnan(corr).any():
                        warnings.warn(
                            "one of the measurements along the repeat_dim has nans"
                        )

                    else:
                        outs[1][ii] += corr

                extra_index += 1
                if all([yshapes[i] == yshapes[refyvar] for i in range(len(yshapes))]):
                    corr_out = out_i[1 + extra_index]
                    if np.isnan(corr).any():
                        warnings.warn(
                            "one of the measurements along the repeat_dim has nans"
                        )

                    else:
                        outs[1 + extra_index] += corr_out
                extra_index += 1

        if return_samples:
            for k in range(lenx):
                if i == 0:
                    outs[1 + extra_index][k] = out_i[1 + extra_index][k]
                    outs[2 + extra_index][k] = out_i[2 + extra_index][k]
                else:
                    outs[1 + extra_index][k] = np.vstack(
                        [outs[1 + extra_index][k], out_i[1 + extra_index][k]]
                    )
                    outs[2 + extra_index][k] = np.vstack(
                        [outs[2 + extra_index][k], out_i[2 + extra_index][k]]
                    )

        return outs

    def finish_repeated_outs(
        self,
        outs,
        yshapes,
        lenx,
        n_repeats,
        repeat_shape,
        repeat_dims,
        return_corr,
        return_samples,
        output_vars,
        refyvar,
    ):
        """
        Do final operations to output arrays with repeated measurements (e.g. dividing by number of repeats to take mean).

        :param outs: array of outputs of the repeated measurements
        :type outs: array
        :type yshapes: tuple
        :param lenx: number of input quantities
        :type lenx: int
        :param n_repeats: number of repeated measurements
        :type n_repeats: int
        :param repeat_shape: shape along which the measurements are repeated
        :type repeat_shape: tuple
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian, defaults to False
        :type return_Jacobian: bool, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param refyvar: Index of output variable with reference shape (only relevant when output_vars>1; should be output variable with most dimensions; affects things like repeat_dims)
        :type refyvar: int
        :return: finalised array of outputs of the repeated measurements
        :rtype: array
        """
        if not return_corr and output_vars == 1 and not return_samples:
            u_func = np.array(outs, dtype=self.dtype)
        else:
            u_func = np.array(outs[0], dtype=self.dtype)
            print("here",u_func[-1].shape)

        if len(repeat_dims) == 1:
            if output_vars == 1:
                u_func = np.squeeze(np.moveaxis(u_func, 0, repeat_dims[0]))
            else:
                u_funcb = u_func[:]
                u_func = np.empty(output_vars, dtype=object)
                if all([yshapes[i] == yshapes[refyvar] for i in range(len(yshapes))]):
                    for i in range(output_vars):
                        u_func[i] = np.moveaxis(u_funcb[i], 0, repeat_dims[0])
                else:
                    for i in range(output_vars):
                        u_func[i] = np.empty(yshapes[i], dtype=self.dtype)
                        if len(yshapes[i]) == len(u_funcb[i].shape):
                            repeat_dim_corr = repeat_dims[0]
                            for idim, dim in enumerate(yshapes[refyvar]):
                                if dim not in yshapes[i] and idim < repeat_dims[0]:
                                    repeat_dim_corr -= 1
                            u_func[i] = np.moveaxis(u_funcb[i], 0, repeat_dim_corr)
                        elif len(yshapes[i]) == 2 and len(u_funcb[i].shape) == 1:
                            for ii in range(yshapes[i][0]):
                                for iii in range(yshapes[i][1]):
                                    u_func[i][ii, iii] = u_funcb[i][iii][ii]
                        elif len(yshapes[i]) == 2 and len(u_funcb[i].shape) == 3:
                            if yshapes[i][0] == u_funcb[i].shape[0]:
                                for ii in range(yshapes[i][0]):
                                    for iii in range(yshapes[i][1]):
                                        u_func[i][ii, iii] = u_funcb[i][ii][0][iii]
                        else:
                            print(yshapes, i, u_funcb[i].shape)
                            raise ValueError(
                                "punpy.mc_propagation: this shape is not supported"
                            )

        else:
            if output_vars == 1:
                u_func = u_func.reshape(repeat_shape + u_func[0].shape)
                for repeat_dim in repeat_dims:
                    u_func = np.moveaxis(u_func, 0, repeat_dim)
            else:
                u_funcb = u_func[:]
                u_func = np.empty(output_vars, dtype=object)
                for i in range(output_vars):
                    u_func[i] = u_funcb[i].reshape(repeat_shape + (-1,))
                    u_func[i] = np.moveaxis(u_func[i], 0, repeat_dims[0])
                    u_func[i] = np.moveaxis(u_func[i], 0, repeat_dims[1])

        if (output_vars == 1 and u_func.shape != yshapes[0]) or (
            output_vars > 1
            and (u_func[0].shape != yshapes[0] or u_func[1].shape != yshapes[1])
        ):
            print(u_func.shape, u_func[0].shape, yshapes)
            raise ValueError(
                "punpy.mc_propagation: The shape of the uncertainties does not match the shape"
                "of the measurand. This is likely a problem with combining"
                "repeated measurements (repeat_dims keyword)."
            )

        if not return_corr and not return_samples:
            return u_func

        else:
            n_masked = len([uf for uf in u_func.T if np.isnan(uf).any()])
            outs[0] = u_func
            if return_corr:
                outs[1] = outs[1] / (n_repeats - n_masked)
                if output_vars == 1:
                    if not cm.isPD(outs[1]):
                        outs[1] = cm.nearestPD_cholesky(
                            outs[1], corr=True, return_cholesky=False
                        )

                else:
                    for j in range(output_vars):
                        if not cm.isPD(outs[1][j]):
                            outs[1][j] = cm.nearestPD_cholesky(
                                outs[1][j], corr=True, return_cholesky=False
                            )

                    if all(
                        [yshapes[i] == yshapes[refyvar] for i in range(len(yshapes))]
                    ):
                        outs[2] = outs[2] / (n_repeats - n_masked)
                        if not cm.isPD(outs[2]):
                            outs[2] = cm.nearestPD_cholesky(
                                outs[2], corr=True, return_cholesky=False
                            )

        return outs

    def run_samples(
        self,
        func,
        MC_x,
        output_vars=1,
        start=None,
        end=None,
        sli=None,
        allow_some_nans=True,
    ):
        """
        process all the MC samples of input quantities through the measurand function

        :param func: measurement function
        :type func: function
        :param MC_x: MC-generated samples of input quantities
        :type MC_x: array[array]
        :param yshapes: shape of the measurand(s)
        :type yshapes: tuple or list of tuples
        :return: MC sample of measurand
        :rtype: array[array]
        """
        if (start is not None) or (end is not None):
            sli = slice(start, end)
            indices = range(*sli.indices(self.MCsteps))
        else:
            sli = slice(sli)
            indices = range(*sli.indices(self.MCsteps))

        if self.parallel_cores == 0:
            if self.MCdimlast:
                MC_y = np.moveaxis(
                    func(*[np.moveaxis(dat[sli], 0, -1) for dat in MC_x]), -1, 0
                )
            else:
                MC_y = func(*[x[sli] for x in MC_x])

        elif self.parallel_cores == 1:
            MC_y = list(map(func, *[x[sli] for x in MC_x]))

        else:
            MC_x2 = np.empty(len(indices), dtype=object)
            for i, index in enumerate(indices):
                MC_x2[i] = [MC_x[j][index, ...] for j in range(len(MC_x))]
            MC_y = self.pool.starmap(func, MC_x2)

        # for i in range(len(indices)):
        #     print(MC_y[i])
        #     print([MC_x[j][i] for j in range(len(MC_x))])
        #     print(output_vars,MC_y[i].shape)
        #     print([np.any(np.isfinite(MC_y[i][ivar])) for ivar in range(output_vars)])
        #     print(np.all([np.any(np.isfinite(MC_y[i][ivar])) for ivar in range(output_vars)]),)
        #     print(np.isfinite(MC_y[i]))


        if output_vars==1:
            if allow_some_nans:
                MC_y = np.array([MC_y[i] for i in range(len(indices)) if np.any(np.isfinite(MC_y[i]))],dtype=self.dtype)
            else:
                MC_y = np.array([MC_y[i] for i in range(len(indices)) if np.all(np.isfinite(MC_y[i]))],dtype=self.dtype)
        else:
            if allow_some_nans:
                try:
                    MC_y = np.array([MC_y[i] for i in range(len(indices)) if np.all([np.any(np.isfinite(MC_y[i][ivar])) for ivar in range(output_vars)])],dtype=self.dtype)
                except:
                    MC_y = np.array([MC_y[i] for i in range(len(indices)) if np.all([np.any(np.isfinite(MC_y[i][ivar])) for ivar in range(output_vars)])],dtype=object)

            else:
                try:
                    MC_y = np.array([MC_y[i] for i in range(len(indices)) if np.all([np.all(np.isfinite(MC_y[i][ivar])) for ivar in range(output_vars)])],dtype=self.dtype)
                except:
                    MC_y = np.array([MC_y[i] for i in range(len(indices)) if np.all([np.all(np.isfinite(MC_y[i][ivar])) for ivar in range(output_vars)])],dtype=object)

        if len(MC_y)<len(indices):
            if allow_some_nans:
                print(
                    "%s of the %s MC samples were not processed correctly (contained only nans) and will be ignored in the punpy output"
                    % (len(indices)-len(MC_y),len(indices))
                )

            else:
                print(
                    "%s of the %s MC samples were not processed correctly (contained some nans) and will be ignored in the punpy output"
                    % (len(indices)-len(MC_y),len(indices))
                )

        if self.verbose:
            print(
                "samples propagated (%s s since creation of prop object)"
                % (time.time() - self.starttime)
            )

        return MC_y

    def combine_samples(self, MC_samples):
        return np.concatenate(MC_samples, axis=0)

    def process_samples(
        self,
        MC_x,
        MC_y,
        return_corr=False,
        return_samples=False,
        yshapes=None,
        corr_dims=-99,
        fixed_corr=None,
        PD_corr=True,
        output_vars=1,
    ):
        """
        Run the MC-generated samples of input quantities through the measurement function and calculate
        correlation matrix if required.

        :param MC_x: MC-generated samples of input quantities
        :type MC_x: array[array]
        :param MC_y: MC sample of measurand
        :type MC_y: array[array]
        :param return_corr: set to True to return correlation matrix of measurand
        :type return_corr: bool
        :param return_samples: set to True to return generated samples
        :type return_samples: bool

        :param corr_dims: set to positive integer to select the axis used in the correla[ tion matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_dims: integer, optional
        :param fixed_corr: correlation matrix to be copied without changing, defaults to None (correlation matrix is calculated rather than copied)
        :type fixed_corr: array
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """

        # if hasattr(MC_y[0,0], '__len__'):
        #     print(yshape,np.array(MC_y[0,0]).shape,np.array(MC_y[1,0]).shape,np.array(MC_y[2,0]).shape,np.array(MC_y[3,0]).shape)
        if output_vars == 1:
            u_func = np.std(MC_y, axis=0, dtype=self.dtype)

        else:
            complex_shapes = True
            if yshapes is None:
                complex_shapes = False
            elif all([yshapes[i] == yshapes[0] for i in range(len(yshapes))]):
                complex_shapes = False

            if complex_shapes:
                MC_y2 = np.empty(output_vars, dtype=object)
                u_func = np.empty(output_vars, dtype=object)

                for i in range(output_vars):
                    MC_y2[i] = np.empty((self.MCsteps,) + yshapes[i])
                    for j in range(self.MCsteps):
                        MC_y2[i][j] = MC_y[j, i]
                    u_func[i] = np.std(np.array(MC_y2[i]), axis=0, dtype=self.dtype)

            else:
                u_func = np.std(MC_y, axis=0, dtype=self.dtype)

        if self.verbose:
            print(
                "std calculated (%s s since creation of prop object)"
                % (time.time() - self.starttime)
            )

        if not return_corr:
            if return_samples:
                return u_func, MC_y, MC_x
            else:
                return u_func
        else:
            if output_vars == 1:
                if fixed_corr is None:
                    corr_y = cm.calculate_corr(MC_y, corr_dims,PD_corr,dtype=self.dtype)
                else:
                    corr_y = fixed_corr

                if self.verbose:
                    print(
                        "corr calculated (%s s since creation of prop object)"
                        % (time.time() - self.starttime)
                    )

                if return_samples:
                    return u_func, corr_y, MC_y, MC_x
                else:
                    return u_func, corr_y

            else:
                # create an empty arrays and then populate it with the correlation matrix for each output parameter individually
                corr_ys = np.empty(output_vars, dtype=object)
                for i in range(output_vars):
                    if fixed_corr is None:
                        if complex_shapes:
                            corr_ys[i] = cm.calculate_corr(MC_y2[i], corr_dims, PD_corr, self.dtype)
                        else:
                            corr_ys[i] = cm.calculate_corr(
                                MC_y[:, i], corr_dims, PD_corr, self.dtype)

                    else:
                        corr_ys[i] = fixed_corr

                if complex_shapes:
                    corr_out = None
                else:
                    # calculate correlation matrix between the different outputs produced by the measurement function.
                    MC_y2 = MC_y.reshape((self.MCsteps, output_vars, -1))

                    corr_out = np.mean(
                        [
                            cm.calculate_corr(MC_y2[:, :, i],PD_corr=PD_corr,dtype=self.dtype)
                            for i in range(len(MC_y2[0, 0]))
                        ],
                        axis=0,
                    )

                if self.verbose:
                    print(
                        "corr calculated for output_var>1 (%s s since creation of prop object)"
                        % (time.time() - self.starttime)
                    )

                if return_samples:
                    return u_func, corr_ys, corr_out, MC_y, MC_x
                else:
                    return u_func, corr_ys, corr_out



"""
Tests for mc propagation class
"""

import unittest

import numpy as np
import numpy.testing as npt

import punpy.utilities.utilities as util

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "14/4/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def function(x1, x2):
    return x1 ** 2 - 10 * x2 + 30.0


x1 = np.ones(200) * 10
x2 = np.ones(200) * 30
x1err = np.ones(200)
x2err = 2 * np.ones(200)

xs = [x1, x2]
xerrs = [x1err, x2err]

# below, the higher order Taylor expansion terms have been taken into account, and amount to 2.
yerr_uncorr = 802 ** 0.5 * np.ones(200)
yerr_corr = 2 ** 0.5 * np.ones(200)


def functionb(x1, x2):
    return 2 * x1 - x2


x1b = np.ones((20, 30)) * 50
x2b = np.ones((20, 30)) * 30
x1errb = np.ones((20, 30))
x2errb = 2 * np.ones((20, 30))

xsb = np.array([x1b, x2b])
xerrsb = np.array([x1errb, x2errb])

yerr_uncorrb = 8 ** 0.5 * np.ones((20, 30))
yerr_corrb = np.zeros((20, 30))


def functionc(x1, x2, x3):
    return x1 + 4 * x2 - 2 * x3


x1c = np.ones(200) * 10
x2c = np.ones(200) * 10
x3c = np.ones(200) * 10

x1errc = 12 * np.ones(200)
x2errc = 5 * np.ones(200)
x3errc = np.zeros(200)

xsc = np.array([x1c, x2c, x3c])
xerrsc = np.array([x1errc, x2errc, x3errc])
corr_c = np.array([[1, 0.9999999, 0], [0.99999999, 1.0, 0], [0.0, 0.0, 1.0]])
yerr_uncorrc = 544 ** 0.5 * np.ones(200)
yerr_corrc = 1024 ** 0.5 * np.ones(200)


def functiond(x1, x2):
    return 2 * x1 - x2, 2 * x1 + x2


x1d = np.ones((20, 3, 4)) * 50
x2d = np.ones((20, 3, 4)) * 30
x1errd = np.ones((20, 3, 4))
x2errd = 2 * np.ones((20, 3, 4))

xsd = [x1d, x2d]
xerrsd = [x1errd, x2errd]
corr_d = np.ones(
    (2, 2)
)  # np.array([[1,0.9999999,0.9999999],[0.99999999,1.,0.99999999],[0.9999999,0.9999999,1.]])

yerr_uncorrd = [
    np.array(8 ** 0.5 * np.ones((20, 3, 4))),
    np.array(8 ** 0.5 * np.ones((20, 3, 4))),
]
yerr_corrd = [np.zeros((20, 3, 4)), 16 ** 0.5 * np.ones((20, 3, 4))]


class TestUtilities(unittest.TestCase):
    """
    Class for unit tests
    """

    def test_select_repeated_x(self):
        xsb2 = np.array([x1b, 1.0])
        xerrsb2 = np.array([x1errb, 2.0])

        out = util.select_repeated_x(xsb2, xerrsb2, [False, True], 0, [0], x1b.shape)
        npt.assert_allclose(out[0][0], x1b[0], atol=0.06)
        npt.assert_allclose(out[0][1], 1, atol=0.06)
        npt.assert_allclose(out[1][0], x1errb[0], atol=0.06)
        npt.assert_allclose(out[1][1], 2.0, atol=0.06)


if __name__ == "__main__":
    unittest.main()



from punpy.digital_effects_table.measurement_function import MeasurementFunction
from punpy.lpu.lpu_propagation import LPUPropagation
from punpy.mc.mc_propagation import MCPropagation
from punpy.utilities.correlation_forms import *
from ._version import __version__

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "01/03/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"
